package com.zell_mbc.publicartexplorer

import android.content.Context
import android.content.Intent
import android.graphics.Canvas
import android.graphics.drawable.Drawable
import android.os.Build
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.text.ClickableText
import androidx.compose.material3.Button
import androidx.compose.material3.CircularProgressIndicator
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Text
import androidx.compose.material3.TextField
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.LinkAnnotation
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.TextLinkStyles
import androidx.compose.ui.text.buildAnnotatedString
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.text.style.TextDecoration
import androidx.compose.ui.text.withLink
import androidx.compose.ui.text.withStyle
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.core.content.ContextCompat
import coil3.compose.AsyncImage
import coil3.compose.SubcomposeAsyncImage
import com.zell_mbc.publicartexplorer.ui.theme.link
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import java.nio.charset.StandardCharsets
import okhttp3.OkHttpClient
import okhttp3.Request
import org.json.JSONObject
import java.net.URLEncoder
import java.security.MessageDigest
import androidx.core.graphics.createBitmap
import androidx.core.net.toUri
import androidx.core.graphics.drawable.toDrawable
import com.zell_mbc.publicartexplorer.data.ViewModel
import java.net.URL
import java.util.Locale
import kotlin.text.isNotEmpty

@Composable
fun AddressInput(onSearch: (String) -> Unit) {
    var address by remember { mutableStateOf("") }

    Column(modifier = Modifier.padding(16.dp)) {
        TextField(
            value = address,
            onValueChange = { address = it },
            label = { Text("Enter address") },
            singleLine = true,
            modifier = Modifier.fillMaxWidth()
        )

        Spacer(modifier = Modifier.height(8.dp))

        Button(
            onClick = { onSearch(address) },
            enabled = address.isNotBlank(),
            modifier = Modifier.align(Alignment.CenterHorizontally)
        ) {
            Text("Search")
        }
    }
}


@Composable
fun ClickableString(preText: String, link: String?, postText:String = "") {
    var s = createClickableUrl(link)
    if (s.isNotEmpty())
        Row {
            Text(preText)
            Text(s)
            if (postText.isNotEmpty()) Text(postText)
            Text("")
        }
}

@Composable
fun createClickableUrl(url: String?, linkText: String = ""): AnnotatedString {
    val linkUrl = url ?: ""

    return buildAnnotatedString {
        withLink(
            LinkAnnotation.Url(
                url = linkUrl,
                styles = TextLinkStyles(
                    style = SpanStyle(
                        color = MaterialTheme.colorScheme.link,
                        textDecoration = TextDecoration.Underline
                    )
                )
            )
        ) {
           if (linkText.isNotEmpty()) append(linkText) else append(url)
        }
    }
}


fun createWikipediaUrl(s: String): String {
    if (s.isEmpty()) return ""

    val (lang, title) = s.split(":", limit = 2)
    val encodedTitle = URLEncoder.encode(title, StandardCharsets.UTF_8.toString())
        .replace("+", "_") // Wikipedia expects underscores for spaces
    return "https://$lang.wikipedia.org/wiki/$encodedTitle"
}

@Composable
fun PanoramaxImage(panoramaxId: String) {
    //val panoramaxId = s //"cafb0ec8-51dd-43ac-836c-8cd1f7cb8725"
    val imageUrl = "https://api.panoramax.xyz/api/pictures/$panoramaxId/sd.jpg"
    Box() {
        SubcomposeAsyncImage(model = imageUrl, loading = { CircularProgressIndicator() }, contentDescription = null,)
        Text(text = "© Panoramax", color = Color.White, fontSize = 12.sp, modifier = Modifier.padding(end = 8.dp).align(Alignment.BottomEnd), textAlign = TextAlign.End)
    }
    Text("")
}

@Composable
fun MapillaryImage(
    imageId: String,
    accessToken: String,
    modifier: Modifier = Modifier
) {
    val context = LocalContext.current
    var imageUrl by remember { mutableStateOf<String?>(null) }
    var error by remember { mutableStateOf<String?>(null) }

    // Fetch the image URL from Mapillary API
    LaunchedEffect(imageId, accessToken) {
        try {
            val apiUrl = "https://graph.mapillary.com/$imageId?access_token=$accessToken&fields=thumb_2048_url"
            val json = withContext(Dispatchers.IO) {
                URL(apiUrl).readText()
            }
            val thumbUrl = JSONObject(json).optString("thumb_2048_url")
            imageUrl = thumbUrl
        } catch (e: Exception) {
            error = context.getString(R.string.failedToLoadImage) + ": ${e.message}"
        }
    }

    Box(modifier = modifier) {
        when {
            error != null -> Text(error ?: stringResource(R.string.unknownError), color = MaterialTheme.colorScheme.error)
            imageUrl != null -> {
                AsyncImage(model = imageUrl, contentDescription = "Mapillary Image", modifier = Modifier.fillMaxWidth().height(200.dp))
                Text(text = "© Mapillary", color = Color.White, fontSize = 12.sp, modifier = Modifier.padding(end = 8.dp).align(Alignment.BottomEnd), textAlign = TextAlign.End)
            }
            else -> CircularProgressIndicator()
        }
    }
    Text("")
}



@Composable
fun ClickableEmail(email: String) {
    val context = LocalContext.current
    val annotatedString = buildAnnotatedString {
        // Add the email as an annotated string
        pushStringAnnotation(tag = "EMAIL", annotation = email)
        withStyle(style = SpanStyle(color = MaterialTheme.colorScheme.link)) { // blue color
            append(email)
        }
        pop()
    }

    ClickableText(
        text = annotatedString,
        onClick = { offset: Int ->
            annotatedString.getStringAnnotations(tag = "EMAIL", start = offset, end = offset)
                .firstOrNull()?.let { annotation ->
                    val intent = Intent(Intent.ACTION_SENDTO).apply {
                        data = "mailto:${annotation.item}".toUri()
                    }
                    context.startActivity(intent)
                }
        }    )
}

suspend fun getDirectImageUrl(filePageUrl: String): String? = withContext(Dispatchers.IO) {
    val fileName = filePageUrl.substringAfterLast(":").replace("_", " ")
    val encodedFileName = URLEncoder.encode(fileName, "UTF-8")
    val apiUrl = "https://commons.wikimedia.org/w/api.php?action=query&titles=File:$encodedFileName&prop=imageinfo&iiprop=url&format=json"

    try {
        val client = OkHttpClient()
        val request = Request.Builder().url(apiUrl).build()
        val response = client.newCall(request).execute()
        val jsonString = response.body?.string() ?: return@withContext null
        val json = JSONObject(jsonString)
        val pages = json.getJSONObject("query").getJSONObject("pages")
        val page = pages.keys().asSequence().map { pages.getJSONObject(it) }.firstOrNull() ?: return@withContext null
        val imageInfo = page.getJSONArray("imageinfo").getJSONObject(0)
        imageInfo.getString("url")
    } catch (_: Exception) {
        null
    }
}

fun getPinWithIcon(context: Context, pinResId: Int, iconResId: Int): Drawable {
    val pinDrawable = ContextCompat.getDrawable(context, pinResId)!!
    val iconDrawable = ContextCompat.getDrawable(context, iconResId)!!

    // Convert both to bitmaps
    val width = pinDrawable.intrinsicWidth
    val height = pinDrawable.intrinsicHeight

    val bitmap = createBitmap(width, height)
    val canvas = Canvas(bitmap)

    // Draw the pin as background
    pinDrawable.setBounds(0, 0, width, height)
    pinDrawable.draw(canvas)

    // Draw the icon centered on the pin
    val iconSize = (width * 0.5).toInt() // Adjust size as needed
    val left = (width - iconSize) / 2
    val top = (height - iconSize) / 2
    iconDrawable.setBounds(left, top, left + iconSize, top + iconSize)
    iconDrawable.draw(canvas)

    return bitmap.toDrawable(context.resources)
}

fun capitalize(input: String) = input.replaceFirstChar { it.titlecase(Locale.getDefault()) }

fun downloadArtworkList(context: Context, viewModel: ViewModel) {
    if (viewModel.artworks.isEmpty()) return

    val sb = StringBuilder()
    val separator = ","
    var line = "longitude,latitude,type" + System.lineSeparator()
    sb.append(line)

    for (item in viewModel.artworks) {
        val type = item.tags["artwork_type"] ?: ""
        val name =  item.tags["name"] ?: "No name"
        line = item.lon.toString() + separator + item.lat.toString() + separator + type + separator + name + System.lineSeparator()
        sb.append(line)
    }

    val sendIntent = Intent().apply {
        action = Intent.ACTION_SEND
        putExtra(Intent.EXTRA_TEXT, sb.toString())
        type = "text/plain"
    }

    val shareIntent = Intent.createChooser(sendIntent, "Share via")
    context.startActivity(shareIntent)
}

fun currentLocale(context: Context): String {
    val currentLocale = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) { context.resources.configuration.locales.get(0) }
    else {
        @Suppress("DEPRECATION")
        context.resources.configuration.locale
    }
    return currentLocale.language // e.g., "en", "fr", "de"
}

// Iterate over language tags and return either the first one or the entry for the current language
fun getLocalValue(key: String, tags: MutableMap<String, String>, context: Context ): String {
    val currentLanguage = currentLocale(context)
    var first = ""
    var local = ""
    for (tag in tags.entries.reversed()) {
        val t = tag.key
        if (t.contains(key)) {
            if (t.contains(":$currentLanguage")) local = tag.value
            else first  = tag.value
        }
    }
    return if (local.isEmpty()) first else local
}
