package com.zell_mbc.publicartexplorer

import android.content.Context
import org.osmdroid.views.MapView
import org.osmdroid.views.overlay.ScaleBarOverlay
import android.graphics.Color
import android.graphics.Color.BLUE
import android.util.DisplayMetrics
import android.util.Log
import android.widget.Toast
import com.zell_mbc.publicartexplorer.data.ViewModel
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import okhttp3.OkHttpClient
import okhttp3.Request
import org.json.JSONObject
import org.osmdroid.util.GeoPoint
import org.osmdroid.views.overlay.Polygon
import java.io.IOException
import kotlin.math.log2
import kotlin.math.cos
import kotlin.text.contains

fun drawSearchCircle(mapView: MapView?, lat: Double, lon: Double, radiusMeters: Double): Polygon {
    val circle = Polygon()
    circle.points = Polygon.pointsAsCircle(GeoPoint(lat, lon), radiusMeters)
    circle.strokeColor = BLUE // Outline color
    circle.fillColor = Color.argb(50, 0, 0, 255)
    circle.strokeWidth = 4f
    circle.infoWindow = null // No popup on tap
    mapView?.overlays?.add(circle)
    mapView?.invalidate()
    return circle
}

fun addScaleBar(mapView: MapView) {
    val scaleBar = ScaleBarOverlay(mapView)
    scaleBar.setCentred(true)

    // Optional: customize appearance
    scaleBar.setScaleBarOffset(mapView.width / 2, mapView.height - 30)
    scaleBar.textPaint.color = Color.BLACK
    scaleBar.barPaint.color = Color.BLACK

    scaleBar.setScaleBarOffset(300, mapView.height - 200)

    // Add to map
    mapView.overlays.add(scaleBar)
    //mapView.invalidate() // refresh the map
}

private const val EARTH_CIRCUMFERENCE = 40075016.686
private const val TILE_SIZE = 256 // OSM standard tile size in pixels

/**
 * Calculates the OSMdroid zoom level for a given ground distance (in meters),
 * latitude (in degrees), and device's actual screen DPI.
 *
 * @param context The Android context.
 * @param meters The ground distance you want to represent (e.g. 1000 for 1 km).
 * @param latitude The latitude at the map center (in degrees).
 * @param scaleBarPixelLength The desired length of the scale bar in physical pixels (default 100px).
 * @return The zoom level as a Double.
 */
fun metersToOsmZoom(
    context: Context,
    meters: Double,
    latitude: Double,
    scaleBarLengthMm: Double = 25.0
): Double {
    val metrics: DisplayMetrics = context.resources.displayMetrics
    val dpi: Float = metrics.densityDpi.toFloat()

    // Convert desired scale bar length from mm to inches, then to pixels
    val scaleBarLengthInches = scaleBarLengthMm / 25.4
    val scaleBarPixelLength = scaleBarLengthInches * dpi

    // meters per pixel for the scale bar
    val metersPerPixel = meters / scaleBarPixelLength

    // meters per pixel at zoom level z:
    // mpp = (Earth circumference * cos(latitude)) / (TILE_SIZE * 2^z)
    // Solve for z:
    // z = log2((Earth circumference * cos(latitude)) / (TILE_SIZE * metersPerPixel))
    val cosLat = cos(Math.toRadians(latitude))
    val denominator = TILE_SIZE * metersPerPixel
    val numerator = EARTH_CIRCUMFERENCE * cosLat
    return log2(numerator / denominator)
}

// Retrieve artworks
// Updates viewModel.artworks and viewModel.relations
suspend fun fetchArtworks(viewModel: ViewModel, server: Int, radius: Int, lat: Double, lon: Double, context: Context, showMuseum: Boolean) = withContext(Dispatchers.IO) {
    val apiServer = apiServers[server]
    val searchTerm = if (showMuseum) "\"tourism\"~\"artwork|museum\"" else "\"tourism\"=\"artwork\""
    //val searchTerm = "\"archaeological_site\""

    //viewModel.artworks.clear() // Should I clean the artworks as well?
    viewModel.relations.clear() // Empty relations array

    val query = """
        [out:json];
        (
        node[$searchTerm](around:$radius,$lat,$lon);
        way[$searchTerm](around:$radius,$lat,$lon);
        relation["type"="site"](around:$radius,$lat,$lon);
        );
        out center;
    """.trimIndent()

    val url = "$apiServer?data=${query.replace("\n", "")}"
    val client = OkHttpClient()
    val request = Request.Builder().url(url).build()
    var success = false
    val artworks = mutableListOf<Artwork>()

    try {
        val response = client.newCall(request).execute()
        if (response.code == 200)
            response.body?.string()?.let { jsonString ->
                val json = JSONObject(jsonString)
                val elements = json.getJSONArray("elements")
                for (i in 0 until elements.length()) {
                    val element = elements.getJSONObject(i)
                    val type = element.getString("type")
                    val id = element.getString("id")

                    // Coordinates
                    when (type) {
                        "node" -> {
                            // Nodes have lat/lon directly
                            val latValue = element.optDouble("lat", Double.NaN)
                            val lonValue = element.optDouble("lon", Double.NaN)
                            val latLonPair = if (latValue.isNaN() || lonValue.isNaN()) null else Pair(latValue, lonValue)
                            if (latLonPair == null) continue // Skip if no valid coordinates

                            var nodeType = NodeType.UNKNOWN
                            val tags = element.optJSONObject("tags")
                            val keyValueStore = mutableMapOf<String, String>()
                            if (tags != null) {
                                val keys = tags.keys()
                                while (keys.hasNext()) {
                                    val key = keys.next()
                                    val value = tags.optString(key)
                                    keyValueStore[key] = value
                                }

                                val tourism = try { (tags["tourism"] ?: "").toString() } catch (_: Exception) { "" }
                                val artworkType = try { (tags["artwork_type"] ?: "").toString() } catch (_: Exception) { "" }

                                nodeType = when {
                                    tourism.contains("museum", ignoreCase = true) -> NodeType.MUSEUM
                                    artworkType.contains("statue", ignoreCase = true) -> NodeType.STATUE
                                    artworkType.contains("sculpture", ignoreCase = true) -> NodeType.SCULPTURE
                                    artworkType.contains("graffiti", ignoreCase = true) -> NodeType.GRAFFITI
                                    artworkType.contains("mural", ignoreCase = true) -> NodeType.MURAL
                                    artworkType.contains("relief", ignoreCase = true) -> NodeType.RELIEF
                                    artworkType.contains("installation", ignoreCase = true) -> NodeType.INSTALLATION
                                    else -> NodeType.UNKNOWN
                                }
                            }
                            artworks.add(Artwork(id = id, lat = latLonPair.first, lon = latLonPair.second, tags = keyValueStore, type = nodeType))
                        }

                        "relation" -> {
                            // Capture tags
                            val tags = element.optJSONObject("tags")
                            val tagsKeyValueStore = mutableMapOf<String, String>()
                            if (tags != null) {
                                val keys = tags.keys()
                                while (keys.hasNext()) {
                                    val key = keys.next()
                                    val value = tags.optString(key)
                                    tagsKeyValueStore[key] = value
                                }
                            }
                            // Capture members
                            val memberArray: ArrayList<String> = arrayListOf<String>()
                            val members = element.optJSONArray("members")
                            if (members != null) {
                                for (i in 0 until members.length()) {
                                    val member = members.getJSONObject(i)
                                    val type = member.optString("type")
                                    val role = member.optString("role")
                                    if (type == "node" && role =="member") {
                                        val ref = member.optLong("ref")
                                        memberArray.add(ref.toString())
                                    }
                                }
                            }
                            viewModel.relations.add(Relation(id = id, tags = tagsKeyValueStore, members = memberArray))
                        }
                    }

            }
        }
        else { // != 200
            withContext(Dispatchers.Main) {
                Toast.makeText(context, context.getString(R.string.unableToRetrieveData) + ", Error: " +response.code, Toast.LENGTH_LONG).show()
                //Toast.makeText(context, "Trying " + apiServers[server+1], Toast.LENGTH_LONG).show()
            }
            //fetchArtworks(server+1, radius, lat, lon, context)
        }

        success = true
    }
    catch (e: IOException) {
        success = false
        Log.d("Error", e.toString())
        withContext(Dispatchers.Main) {
            Toast.makeText(context, context.getString(R.string.unableToRetrieveData) + " " +e.message, Toast.LENGTH_LONG).show()
            //Toast.makeText(context, "Trying " + apiServers[server+1], Toast.LENGTH_LONG).show()
        }
    }

    if (success)
        withContext(Dispatchers.Main) {
            if (artworks.isEmpty()) {
                Toast.makeText(context, context.getString(R.string.noArtworksInCircle, artworks.size), Toast.LENGTH_LONG).show()
            } else {
                Toast.makeText(context, context.getString(R.string.nrArtworksInCircle, artworks.size), Toast.LENGTH_LONG).show()
            }
        }

    viewModel.resetArtworkList(artworks)
}
