package com.zell_mbc.publicartexplorer

import android.content.Context
import android.content.SharedPreferences
import android.location.Address
import android.location.Geocoder
import android.widget.Toast
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.size
import androidx.compose.material3.AlertDialog
import androidx.compose.material3.Button
import androidx.compose.material3.CircularProgressIndicator
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.OutlinedButton
import androidx.compose.material3.Text
import androidx.compose.material3.TextField
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.unit.dp
import androidx.core.content.edit
import androidx.datastore.core.IOException
import kotlinx.coroutines.launch
import org.osmdroid.util.GeoPoint
import org.osmdroid.views.MapView


@Composable
fun AddressInput(
    context: Context,
    mapView: MapView?,
    preferences: SharedPreferences,
    onSearchComplete: () -> Unit
) {
    var query by remember { mutableStateOf("") }
    var searchInProgress by remember { mutableStateOf(false) }
    val coroutineScope = rememberCoroutineScope()

    AlertDialog(
        onDismissRequest = { onSearchComplete() },
        title = { Text("Search Address") },
        text = {
            Column {
                TextField(value = query, onValueChange = { query = it }, label = { Text("Address") }, enabled = !searchInProgress)
            }
        },
        confirmButton = {
            Button(onClick = {
                if (!searchInProgress) {
                    searchInProgress = true
                    coroutineScope.launch {
                        try {
                            val geocoder = Geocoder(context)
                            val addresses = geocoder.getFromLocationName(query, 1) // Modern suspend function

                            if (!addresses.isNullOrEmpty()) {
                                val address = addresses[0]
                                val latitude = address.latitude
                                val longitude = address.longitude

                                mapView?.controller?.apply {
                                    setZoom(Constants.SEARCH_ZOOM)
                                    setCenter(GeoPoint(latitude, longitude))
                                }
                                preferences.edit {
                                    putLong(LATITUDE_KEY, latitude.toRawBits())
                                    putLong(LONGITUDE_KEY, longitude.toRawBits())
                                    putLong(ZOOM_KEY, Constants.SEARCH_ZOOM.toRawBits())
                                }
                                onSearchComplete()
                            } else {
                                Toast.makeText(context, context.getString(R.string.noAddressFound), Toast.LENGTH_LONG).show()
                            }
                        } catch (e: IOException) {
                            Toast.makeText(context, context.getString(R.string.unknownError) + e.message, Toast.LENGTH_LONG).show()
                        } finally {
                            searchInProgress = false
                        }
                    }
                }
            },
                enabled = !searchInProgress
            ) {
                if (searchInProgress) CircularProgressIndicator(color = MaterialTheme.colorScheme.onPrimary, modifier = Modifier.size(16.dp))
                else Text(stringResource(R.string.search))
            }
        },
        dismissButton = {
            OutlinedButton(
                onClick = { onSearchComplete() },
                enabled = !searchInProgress
            ) {
                Text("Cancel")
            }
        }
    )
}


// can't yet use the below because it requires Android 14 (API 34)
/*@Composable
fun AddressInput(
    context: Context,
    mapView: MapView?,
    preferences: SharedPreferences,
    onSearchComplete: () -> Unit
) {
    var query by remember { mutableStateOf("") }
    var searchInProgress by remember { mutableStateOf(false) }
    val coroutineScope = rememberCoroutineScope()

    AlertDialog(
        onDismissRequest = { onSearchComplete() },
        title = { Text("Search Address") },
        text = {
            Column {
                TextField(value = query, onValueChange = { query = it }, label = { Text("Address") }, enabled = !searchInProgress)
            }
        },
        confirmButton = {
            Button(onClick = {
                if (!searchInProgress) {
                    searchInProgress = true
                    searchAddress(
                        context = context,
                        query = query,
                        onResult = { addresses ->
                            searchInProgress = false
                            if (!addresses.isNullOrEmpty()) {
                                val address = addresses[0]
                                val latitude = address.latitude
                                val longitude = address.longitude

                                mapView?.controller?.apply {
                                    setZoom(Constants.SEARCH_ZOOM)
                                    setCenter(GeoPoint(latitude, longitude))
                                }
                                preferences.edit {
                                    putLong(LATITUDE_KEY, latitude.toRawBits())
                                    putLong(LONGITUDE_KEY, longitude.toRawBits())
                                    putLong(ZOOM_KEY, Constants.SEARCH_ZOOM.toRawBits())
                                }
                            } else {
                                Toast.makeText(context, "No address found.", Toast.LENGTH_LONG).show()
                            }
                        },
                        onError = { e ->
                            searchInProgress = false
                            Toast.makeText(context, "Error: ${e.message}", Toast.LENGTH_LONG).show()
                        }
                    )
                }
            },
                enabled = !searchInProgress            ) {
                if (searchInProgress) CircularProgressIndicator(color = MaterialTheme.colorScheme.onPrimary, modifier = Modifier.size(16.dp))
                else Text(stringResource(R.string.search))
            }
        },
        dismissButton = {
            OutlinedButton(
                onClick = { onSearchComplete() },
                enabled = !searchInProgress
            ) {
                Text("Cancel")
            }
        }
    )
}

fun searchAddress(
    context: Context,
    query: String,
    onResult: (List<Address>?) -> Unit,
    onError: (Exception) -> Unit
) {
    val geocoder = Geocoder(context)
    try {
        geocoder.getFromLocationName(query, 1,
            object : Geocoder.GeocodeListener {
                override fun onGeocode(addresses: MutableList<Address>) {
                    onResult(addresses)
                }
                override fun onError(errorMessage: String?) {
                    onError(Exception(errorMessage ?: context.getString(R.string.unknownError)))
                }
            }
        )
    } catch (e: Exception) {
        onError(e)
    }
}*/