package com.zell_mbc.publicartexplorer.detailview

import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.padding
import androidx.compose.material3.CircularProgressIndicator
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import coil3.compose.SubcomposeAsyncImage
import java.net.URL
import java.net.URLDecoder
import java.security.MessageDigest
import com.zell_mbc.publicartexplorer.R
import com.zell_mbc.publicartexplorer.capitalize
import com.zell_mbc.publicartexplorer.getDirectImageUrl

@Composable
fun ImageTag(imageUrl: String) {
    if (isMediaViewerUrl(imageUrl)) {
        val fileName = imageUrl.substring(imageUrl.lastIndexOf(":")+1) // extract filename and create new url
        val newUrl = getWikimediaImageUrl(fileName)
        if (newUrl != null ) {
            val attribution = capitalize(extractWikiWord(newUrl) ?: stringResource(R.string.unknown))
            ShowImage(newUrl, attribution)
        }
    }
    else {
        if (imageUrl.contains("https://commons.wikimedia.org/wiki/File:"))
            WikimediaImage(imageUrl)
        else
            if (imageUrl.contains("http")) { // Already a proper URL
                val attribution = capitalize(extractWikiWord(imageUrl) ?: stringResource(R.string.unknown))
                Box() {
                    SubcomposeAsyncImage(
                        model = imageUrl,
                        loading = { CircularProgressIndicator() },
                        contentDescription = null,
                    )
                    Text(
                        text = "© $attribution",
                        color = Color.White,
                        fontSize = 12.sp,
                        modifier = Modifier.padding(end = 8.dp).align(Alignment.BottomEnd),
                        textAlign = TextAlign.End
                    )
                }
                Text("")
            } else
                WikimediaImage(imageUrl)
    }
}


fun isMediaViewerUrl(urlString: String): Boolean {
    return try {
        val url = URL(urlString)
        // Accept Wikipedia language subdomains and Commons
        val hostIsWiki = Regex("""^([a-z]{2,3}\.wikipedia\.org|commons\.wikimedia\.org)$""").matches(url.host)
        val pathIsWikiOrCommons = url.path.startsWith("/wiki/")
        val frag = url.ref ?: ""
        val isMediaViewer = frag.contains("/media/File:") || frag.contains("/media/Datei:")
        hostIsWiki && pathIsWikiOrCommons && isMediaViewer
    } catch (e: Exception) {
        false
    }
}


// Wikipedia "media viewer" URL
// Call with file name
fun getWikimediaImageUrl(inputUrl: String): String? {
    // Try to extract file name from typical Wikipedia/Commons URLs
    val fileName = when {
        // Media viewer fragment (Wikipedia or Commons): #/media/File:... or #/media/Datei:...
        inputUrl.contains("#/media/File:") -> inputUrl.substringAfter("#/media/File:")
        inputUrl.contains("#/media/Datei:") -> inputUrl.substringAfter("#/media/Datei:")
        // File page URL: /wiki/File:... or /wiki/Datei:...
        inputUrl.contains("/wiki/File:") -> inputUrl.substringAfter("/wiki/File:")
        inputUrl.contains("/wiki/Datei:") -> inputUrl.substringAfter("/wiki/Datei:")
        // Direct filename as input
        inputUrl.endsWith(".jpg", true) || inputUrl.endsWith(".png", true) -> inputUrl
        else -> null
    }?.let { URLDecoder.decode(it, "UTF-8") } ?: return null

    // Compute MD5 hash of the file name (case sensitive, spaces and underscores as in URL)
    val md5 = MessageDigest.getInstance("MD5")
    val hash = md5.digest(fileName.toByteArray()).joinToString("") { "%02x".format(it) }
    val first = hash.substring(0, 1)
    val firstTwo = hash.substring(0, 2)

    return "https://upload.wikimedia.org/wikipedia/commons/$first/$firstTwo/$fileName"
}


@Composable
fun ShowImage(imageUrl: String, attribution: String) {
    Box() {
        SubcomposeAsyncImage(model = imageUrl, loading = { CircularProgressIndicator() }, contentDescription = null,)
        Text(text = "© $attribution", color = Color.White, fontSize = 12.sp, modifier = Modifier.padding(end = 8.dp).align(Alignment.BottomEnd), textAlign = TextAlign.End)
    }
    Text("")
}


fun extractWikiWord(input: String): String? {
    val regex = Regex("""\b(wikipedia|wikimedia)\b""", RegexOption.IGNORE_CASE)
    return regex.find(input)?.value?.lowercase()
}

@Composable
fun WikimediaImage(filePageUrl: String) {
    var realImageUrl by remember { mutableStateOf<String?>(null) }

    // Launch a coroutine to fetch the real image URL when filePageUrl changes
    LaunchedEffect(filePageUrl) {
        realImageUrl = getDirectImageUrl(filePageUrl)
    }

    // Show the image when the real URL is available
    if (realImageUrl != null) {
        Box() {
            SubcomposeAsyncImage(model = realImageUrl, loading = { CircularProgressIndicator() }, contentDescription = null, error = { Text(stringResource(
                R.string.unableToRetrieveData)) })
            Text(text = "© Wikimedia", color = Color.White, fontSize = 12.sp, modifier = Modifier.padding(end = 8.dp).align(Alignment.BottomEnd), textAlign = TextAlign.End)
        }
        Text("")
    } else {
        // Optional: show a loading indicator or placeholder
        CircularProgressIndicator()
    }
}
