package com.zell_mbc.publicartexplorer.screens

import android.Manifest
import android.content.Context
import android.content.pm.PackageManager
import android.graphics.Bitmap
import android.graphics.Canvas
import android.graphics.Paint
import android.graphics.PorterDuff
import android.graphics.PorterDuffColorFilter
import android.graphics.Rect
import android.graphics.drawable.BitmapDrawable
import android.graphics.drawable.Drawable
import android.location.Location
import android.location.LocationManager
import android.location.LocationRequest
import android.os.Build
import android.os.CancellationSignal
import android.os.Handler
import android.os.Looper
import android.preference.PreferenceManager
import android.view.ViewTreeObserver
import android.widget.Toast
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.rememberScrollState
import androidx.compose.foundation.verticalScroll
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Refresh
import androidx.compose.material3.AlertDialog
import androidx.compose.material3.FloatingActionButton
import androidx.compose.material3.Icon
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Text
import androidx.compose.material3.TextButton
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.toArgb
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.res.painterResource
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.compose.ui.viewinterop.AndroidView
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import androidx.core.content.edit
import com.google.accompanist.permissions.ExperimentalPermissionsApi
import com.google.accompanist.permissions.rememberMultiplePermissionsState
import com.zell_mbc.publicartexplorer.AddressInput
import com.zell_mbc.publicartexplorer.Artwork
import com.zell_mbc.publicartexplorer.RADIUS_KEY
import com.zell_mbc.publicartexplorer.Constants
import com.zell_mbc.publicartexplorer.Constants.DEFAULT_ZOOM
import com.zell_mbc.publicartexplorer.EXPERT_MODE_KEY
import com.zell_mbc.publicartexplorer.LATITUDE_KEY
import com.zell_mbc.publicartexplorer.LAUNCH_AT_CURRENT_POSITION_KEY
import com.zell_mbc.publicartexplorer.LONGITUDE_KEY
import com.zell_mbc.publicartexplorer.MainActivity.Companion.firstStart
import com.zell_mbc.publicartexplorer.NodeType
import com.zell_mbc.publicartexplorer.R
import com.zell_mbc.publicartexplorer.SHOW_MUSEUM_KEY
import com.zell_mbc.publicartexplorer.ZOOM_KEY
import com.zell_mbc.publicartexplorer.data.ViewModel
import com.zell_mbc.publicartexplorer.detailview.ShowArtwork
import com.zell_mbc.publicartexplorer.detailview.ShowRawData
import com.zell_mbc.publicartexplorer.downloadArtworkList
import com.zell_mbc.publicartexplorer.drawSearchCircle
import com.zell_mbc.publicartexplorer.fetchArtworks
import com.zell_mbc.publicartexplorer.mapView
import kotlinx.coroutines.launch
import org.osmdroid.config.Configuration
import org.osmdroid.events.MapListener
import org.osmdroid.events.ScrollEvent
import org.osmdroid.events.ZoomEvent
import org.osmdroid.tileprovider.tilesource.TileSourceFactory
import org.osmdroid.util.GeoPoint
import org.osmdroid.views.MapView
import org.osmdroid.views.overlay.Marker
import org.osmdroid.views.overlay.Polygon
import org.osmdroid.views.overlay.ScaleBarOverlay
import org.osmdroid.views.overlay.mylocation.GpsMyLocationProvider
import org.osmdroid.views.overlay.mylocation.MyLocationNewOverlay
import java.util.concurrent.Executors


@OptIn(ExperimentalPermissionsApi::class)
@Composable
fun MapScreen(viewModel: ViewModel) {
    val context = LocalContext.current
    val preferences = PreferenceManager.getDefaultSharedPreferences(context)

    val radiusPreference = preferences.getInt(RADIUS_KEY, Constants.DEFAULT_RADIUS)
    val expertModePreference = preferences.getBoolean(EXPERT_MODE_KEY, Constants.DEFAULT_EXPERT_MODE)
    val showMuseumPreference = preferences.getBoolean(SHOW_MUSEUM_KEY, Constants.DEFAULT_SHOW_MUSEUM)
    val openAppAtCurrentPositionPreference = preferences.getBoolean(LAUNCH_AT_CURRENT_POSITION_KEY, Constants.DEFAULT_LAUNCH_AT_CURRENT_POSITION)

    var latitudePreference = Double.fromBits(preferences.getLong(LATITUDE_KEY, 0L))
    val longitudePreference = Double.fromBits(preferences.getLong(LONGITUDE_KEY, 0L))
    var zoomPreference = Double.fromBits(preferences.getLong(ZOOM_KEY, 0L))
    if (zoomPreference == 0.0) zoomPreference = DEFAULT_ZOOM

    // Preferences
/*    val radius by viewModel.getRadius.collectAsState()
    val expertMode by viewModel.getExpertMode.collectAsState()
    val longitude by viewModel.getLongitude.collectAsState()
    val latitude by viewModel.getLatitude.collectAsState()
    val zoom by viewModel.getZoom.collectAsState()
*/


    // Step 1: Ensure osmdroid config and user agent
    LaunchedEffect(Unit) {
        Configuration.getInstance().load(context, PreferenceManager.getDefaultSharedPreferences(context))
        Configuration.getInstance().userAgentValue = context.packageName
    }

    var circleOverlay: Polygon? by remember { mutableStateOf(null) }
    var artworks = viewModel.artworks // remember { mutableStateOf(listOf<Artwork>()) }
    var selectedArtwork by remember { mutableStateOf<Artwork?>(null) }
    var showRawData  by remember { mutableStateOf<Boolean>(false) }
    var showSearch   by remember { mutableStateOf<Boolean>(false) }
    var longitude by remember { mutableStateOf<Double>(longitudePreference) }
    var latitude by remember { mutableStateOf<Double>(latitudePreference) }
    var zoom by remember { mutableStateOf<Double>(zoomPreference) }
    var expertMode by remember { mutableStateOf<Boolean>(expertModePreference) }
    var showLocateMeFab by remember { mutableStateOf<Boolean>(true) }
    var showMuseum by remember { mutableStateOf<Boolean>(showMuseumPreference) }
    var radius by remember { mutableStateOf<Int>(radiusPreference) }
    var apiServer by remember { mutableStateOf<Int>(0) }
    var openAppAtCurrentPosition by remember { mutableStateOf<Boolean>(openAppAtCurrentPositionPreference) }

    // val loadCurrentPosition by viewModel.getLoadPosition.collectAsState()

    // Request both fine and coarse location permissions
    val permissionsState = rememberMultiplePermissionsState(
        permissions = listOf(
            Manifest.permission.ACCESS_FINE_LOCATION,
            Manifest.permission.ACCESS_COARSE_LOCATION
        )
    )

    LaunchedEffect(Unit) {
        Configuration.getInstance().load(context, PreferenceManager.getDefaultSharedPreferences(context))
        Configuration.getInstance().userAgentValue = context.packageName
        permissionsState.launchMultiplePermissionRequest()
      }

    // When permissions are granted, get location and center map
    // Sets map to current position on first launch, to whatever is saved in Preferences on all subsequent calls
    LaunchedEffect(permissionsState.allPermissionsGranted) {
        showLocateMeFab = true
        //if (permissionsState.allPermissionsGranted) {
            // First start after installation
            // Lat, Lon == 0
            // Set to current location, no matter what openAppAtCurrentPosition says
        if (longitude == 0.0 && latitude == 0.0) {
            zoom = DEFAULT_ZOOM
            //var pos = Location(latitude = 48.0622, longitude = 9.6606)
            getLocation(context) { location ->
                //if (pos != null) {
                    latitude = location.latitude
                    longitude = location.longitude
                    preferences.edit {
                        putLong(LATITUDE_KEY, latitude.toRawBits()).apply()
                        putLong(LONGITUDE_KEY, longitude.toRawBits()).apply()
                    }
                /*else {
                    // If we can't get the current location, point to Oggelshausen, because after all that's where it all started
                    latitude = 48.0622
                    longitude = 9.6606
                }*/
            }
        }
        // Pull values from preferences
        else
            // App restart
            if (firstStart) {
                firstStart = false
                // Show last time location or current location depending on setting
                if (openAppAtCurrentPosition) {
                    getLocation(context) { location ->
                            latitude = location.latitude
                            longitude = location.longitude
                            preferences.edit {
                                putLong(LATITUDE_KEY, latitude.toRawBits()).apply()
                                putLong(LONGITUDE_KEY, longitude.toRawBits()).apply()
                            }
                    }
                }
            }
        // Everything else is a recompose, nothing needs to be done
        val geoPoint = GeoPoint(latitude,longitude)
        mapView?.controller?.setZoom(zoom)
        mapView?.controller?.setCenter(geoPoint)
       // }
    }

    // Optionally, handle UI for denied permissions
    if (!permissionsState.allPermissionsGranted) {
        // Todo : Hide FAB
        showLocateMeFab = false
    }

    // Step 4: Display Map and Markers
    val iconBackground = MaterialTheme.colorScheme.tertiaryContainer.toArgb()
    val iconForeground = MaterialTheme.colorScheme.onTertiaryContainer.toArgb()

    // https://fonts.google.com/icons
    // httpss noun project
    val icon_museum = createCircledIcon(context, R.drawable.icon_museum, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_statue = createCircledIcon(context, R.drawable.icon_statue, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_sculpture = createCircledIcon(context, R.drawable.icon_artwork, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_graffiti = createCircledIcon(context, R.drawable.icon_graffiti, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_mural = createCircledIcon(context, R.drawable.icon_mural, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_relief = createCircledIcon(context, R.drawable.icon_mural, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_installation = createCircledIcon(context, R.drawable.icon_installation, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_unknown = createCircledIcon(context, R.drawable.icon_question_mark, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_fountain = createCircledIcon(context, R.drawable.icon_fountain, backgroundColor = iconBackground, foregroundColor = iconForeground)
    val icon_architecture = createCircledIcon(context, R.drawable.icon_architecture, backgroundColor = iconBackground, foregroundColor = iconForeground)

    Box(modifier = Modifier.fillMaxSize()) {
        AndroidView(
            factory = { ctx ->
                MapView(ctx).apply {
                    mapView = this
                    //addScaleBar(mapView = mapView!!)

                    addMapListener(object : MapListener {
                        private val scrollUpdateHandler = Handler(Looper.getMainLooper())
                        private var scrollUpdateRunnable: Runnable? = null

                        override fun onScroll(event: ScrollEvent?): Boolean {
                            // Cancel any pending callbacks
                            scrollUpdateRunnable?.let { scrollUpdateHandler.removeCallbacks(it) }
                            // Schedule a new callback
                            scrollUpdateRunnable = Runnable {
                                if (permissionsState.allPermissionsGranted) { // If not we are getting bogus values
                                    val center = mapView?.mapCenter as GeoPoint
                                    latitude = center.latitude
                                    longitude = center.longitude
                                    preferences.edit {
                                        putLong(LATITUDE_KEY, center.latitude.toRawBits()).apply()
                                        putLong(LONGITUDE_KEY, center.longitude.toRawBits()).apply()
                                    }
                                }
                            }
                            scrollUpdateHandler.postDelayed(scrollUpdateRunnable!!, 500)
                            return true
                        }

                        private val zoomUpdateHandler = Handler(Looper.getMainLooper())
                        private var zoomUpdateRunnable: Runnable? = null

                        override fun onZoom(event: ZoomEvent?): Boolean {
                            val newZoom = mapView?.zoomLevelDouble
                            if (newZoom != null && newZoom != zoom) {
                                // Cancel any pending update
                                zoomUpdateRunnable?.let { zoomUpdateHandler.removeCallbacks(it) }
                                // Schedule a new update after 500ms of inactivity
                                zoomUpdateRunnable = Runnable {
                                    preferences.edit { putLong(ZOOM_KEY, newZoom.toRawBits()).apply() }
                                }
                                zoomUpdateHandler.postDelayed(zoomUpdateRunnable!!, 500)
                            }
                            return true
                        }
                    })

                    setTileSource(TileSourceFactory.MAPNIK)
                    setMultiTouchControls(true)
                    controller.setCenter(GeoPoint(latitude, longitude))
                    mapView!!.zoomController.setVisibility(org.osmdroid.views.CustomZoomButtonsController.Visibility.NEVER)
                    //val initialZoom = metersToOsmZoom(context, 500.0, latitude)
                    controller.setZoom(zoom)
                }
            },
            update = { mapView ->
                //val center = mapView.mapCenter as GeoPoint
                mapView.overlays.clear()
                artworks.forEach { artwork ->
                    val marker = Marker(mapView).apply {
                        val node_icon = when (artwork.type) {
                        NodeType.MUSEUM -> icon_museum
                        NodeType.STATUE -> icon_statue
                        NodeType.SCULPTURE -> icon_sculpture
                        NodeType.GRAFFITI -> icon_graffiti
                        NodeType.MURAL -> icon_mural
                        NodeType.RELIEF -> icon_relief
                        NodeType.INSTALLATION -> icon_installation
                        NodeType.ARCHITECTURE -> icon_architecture
                        NodeType.FOUNTAIN -> icon_fountain
                        else -> icon_unknown
                    }
                        icon = node_icon
                        position = GeoPoint(artwork.lat, artwork.lon)
                        setAnchor(Marker.ANCHOR_CENTER, Marker.ANCHOR_CENTER)
                        setOnMarkerClickListener { m, _ ->
                            selectedArtwork = artwork
                            true
                        }
                    }
                    mapView.overlays.add(marker)
                }
                //scaleBar.setCentred(true)
                //scaleBar.textPaint.color = android.graphics.Color.BLACK
                //scaleBar.barPaint.color = android.graphics.Color.BLACK

                mapView.invalidate()

                // Wait for layout to position the scale bar
                mapView.viewTreeObserver.addOnGlobalLayoutListener(object : ViewTreeObserver.OnGlobalLayoutListener {
                    override fun onGlobalLayout() {
                        // Remove listener to avoid repeated calls
                        mapView.viewTreeObserver.removeOnGlobalLayoutListener(this)
                        // Now mapView.height is valid
                        val scaleBarOverlay = ScaleBarOverlay(mapView)
                        scaleBarOverlay.setScaleBarOffset(16, mapView.height - 180)
                        mapView.overlays.add(scaleBarOverlay)
                        // Optionally, force a redraw
                        mapView.invalidate()
                    }
                })
            }
          )
        // Copyright overlay
        Text(text = "© OpenStreetMap contributors", color = Color.Black, fontSize = 12.sp,
            modifier = Modifier.align(Alignment.BottomStart).padding(8.dp).background(Color.White.copy(alpha = 0.7f)).padding(horizontal = 8.dp, vertical = 2.dp))

        val coroutineScope = rememberCoroutineScope()
        // Locate me button
        if (showLocateMeFab)
            FloatingActionButton(
                onClick = {
                    // Create and add the location overlay
                    var myLocationOverlay = MyLocationNewOverlay(GpsMyLocationProvider(context), mapView)
                    myLocationOverlay.enableMyLocation()
                    myLocationOverlay.enableFollowLocation()
                    myLocationOverlay.isDrawAccuracyEnabled = true
                    mapView?.overlays?.add(myLocationOverlay)

                    // Check lastLocation first -> Immediate response
                    var pos = getLastLocation(context)
                    if (pos == null) Toast.makeText(context, context.getString(R.string.tryingToRetrieveLocation), Toast.LENGTH_LONG).show()
                    // Move the map to your location when available
                    myLocationOverlay.runOnFirstFix {
                        coroutineScope.launch {
                            mapView?.controller?.animateTo(myLocationOverlay.myLocation)
                            preferences.edit {
                                putLong(LATITUDE_KEY, myLocationOverlay.myLocation.latitude.toRawBits()).apply()
                                putLong(LONGITUDE_KEY, myLocationOverlay.myLocation.longitude.toRawBits()).apply()
                                putLong(ZOOM_KEY, zoom.toRawBits()).apply()
                            }
                        }
                    }
                },
                modifier = Modifier.padding(16.dp).align(Alignment.TopEnd)
            ) {
                Icon(painter = painterResource(id = R.drawable.my_location), contentDescription = "Reset position")
            }

        // ShowCircle button
        var topPos = if (showLocateMeFab) 80 else 16
        FloatingActionButton(
            onClick = {
                //Toast.makeText(context, "circleOverlay: " + circleOverlay, Toast.LENGTH_LONG).show()
                if (circleOverlay == null) {
                    val center = mapView?.mapCenter as GeoPoint
                    circleOverlay = drawSearchCircle(mapView = mapView, lat = center.latitude, lon = center.longitude, radius.toDouble())
                    preferences.edit {
                        putLong(LATITUDE_KEY, center.latitude.toRawBits()).apply()
                        putLong(LONGITUDE_KEY, center.longitude.toRawBits()).apply()
                    }
                }
                else {
                    mapView?.overlays?.remove(circleOverlay)
                    mapView?.invalidate()
                    circleOverlay = null
                }
            },
            modifier = Modifier
                .padding(top = topPos.dp, end = 16.dp)
                .align(Alignment.TopEnd)
        ) {
            if (circleOverlay == null) Icon(painter = painterResource(id = R.drawable.plus_circle), contentDescription = "Show circle")
            else Icon(painter = painterResource(id = R.drawable.minus_circle), contentDescription = "Hide circle")
        }

        // Search button
        topPos = if (showLocateMeFab) 145 else 80
        FloatingActionButton(onClick = { showSearch = true }, modifier = Modifier
            .padding(top = topPos.dp, end = 16.dp)
            .align(Alignment.TopEnd)) {
            Icon(painter = painterResource(id = R.drawable.ic_search), contentDescription = "Search")
           }

        // Download button
        topPos = if (showLocateMeFab) 210 else 145
        if (expertMode && artworks.size > 0)
            FloatingActionButton(onClick = {
               downloadArtworkList(context, viewModel)
            }, modifier = Modifier
                .padding(top = 210.dp, end = 16.dp)
                .align(Alignment.TopEnd)) {
                Icon(painter = painterResource(id = R.drawable.ic_download), contentDescription = "Search")
            }

        // Reload artworks button
        FloatingActionButton(
            onClick = {
                // Draw circle
                coroutineScope.launch {
                    Toast.makeText(context, context.getString(R.string.loadingArtworks, radius), Toast.LENGTH_LONG).show()
                    val center = mapView?.mapCenter as GeoPoint
                    // Can't check size of array here because the Array may still be loading
                    fetchArtworks(viewModel = viewModel, showMuseum = showMuseum, server = apiServer, radius = radius, lat = center.latitude, lon = center.longitude, context = context)
                }
            },
            modifier = Modifier
                .padding(16.dp)
                .align(Alignment.BottomEnd)
        ) {
            Icon(Icons.Filled.Refresh, contentDescription = "Reload Artwork")
        }
        //Text(artworks.size.toString())
    }
    if (selectedArtwork != null) {
        AlertDialog(
            onDismissRequest = { selectedArtwork = null },
            confirmButton = { TextButton(onClick = { selectedArtwork = null }) {  Text(stringResource(R.string.done)) } },
            dismissButton = if (expertMode) { {
                TextButton(onClick = { showRawData = !showRawData }) { Text(if (showRawData) stringResource(R.string.details) else stringResource(R.string.rawData)) }
            } } else null,
            title = { Text(if (!showRawData) (stringResource(R.string.details)  + ":") else (stringResource(R.string.rawData) + ":")) },
            text = {
                Column(modifier = Modifier
                    .fillMaxWidth()
                    .verticalScroll(rememberScrollState())) {
                    if (showRawData) ShowRawData(viewModel, selectedArtwork!!)
                    else ShowArtwork(viewModel, selectedArtwork!!)
                }
            },
        )
    }
    if (showSearch) {
        AddressInput(
            context = context,
            mapView = mapView,
            preferences = preferences,
            onSearchComplete = { showSearch = false }
        )
        }
}

fun createCircledIcon(context: Context, iconResId: Int, sizePx: Int = 96, backgroundColor: Int, foregroundColor: Int): Drawable {
    // Load your icon as a Bitmap
    val iconDrawable = ContextCompat.getDrawable(context, iconResId)!!
    val iconBitmap = Bitmap.createBitmap(sizePx, sizePx, Bitmap.Config.ARGB_8888)
    val iconCanvas = Canvas(iconBitmap)
    iconDrawable.setBounds(0, 0, sizePx, sizePx)
    iconDrawable.draw(iconCanvas)

    // Create a new bitmap for the final marker
    val outputBitmap = Bitmap.createBitmap(sizePx, sizePx, Bitmap.Config.ARGB_8888)
    val canvas = Canvas(outputBitmap)

    // Draw filled circle
    val paint = Paint(Paint.ANTI_ALIAS_FLAG)
    paint.color = backgroundColor
    paint.style = Paint.Style.FILL
    val radius = sizePx / 2f
    canvas.drawCircle(radius, radius, radius, paint)

    // Draw black circumference
    paint.color = android.graphics.Color.BLACK
    paint.style = Paint.Style.STROKE
    paint.strokeWidth = sizePx * 0.04f // 8% of size as border thickness
    canvas.drawCircle(radius, radius, radius - paint.strokeWidth / 2, paint)

    // Draw the icon centered
    val foregroundPaint = Paint(Paint.ANTI_ALIAS_FLAG).apply {
        colorFilter = PorterDuffColorFilter(foregroundColor, PorterDuff.Mode.SRC_IN)
    }
    // Draw the icon centered and tinted
    val iconInset = (sizePx * 0.18f).toInt()
    val iconRect = Rect(iconInset, iconInset, sizePx - iconInset, sizePx - iconInset)
    canvas.drawBitmap(iconBitmap, null, iconRect, foregroundPaint)

    return BitmapDrawable(context.resources, outputBitmap)
}


//fun getLocation( context: Context): Location? {
fun getLocation(context: Context, callback: (Location) -> Unit) {
    var loc: Location? = null

    // Check permissions
    if (ActivityCompat.checkSelfPermission(context, Manifest.permission.ACCESS_FINE_LOCATION) != PackageManager.PERMISSION_GRANTED &&
        ActivityCompat.checkSelfPermission(context, Manifest.permission.ACCESS_COARSE_LOCATION) != PackageManager.PERMISSION_GRANTED ) {
        return
    }

    val locationManager = context.getSystemService(Context.LOCATION_SERVICE) as LocationManager

    // Check if providers are enabled
    val isGpsEnabled = locationManager.isProviderEnabled(LocationManager.GPS_PROVIDER)
    val isNetworkEnabled = locationManager.isProviderEnabled(LocationManager.NETWORK_PROVIDER)

    // No provider enabled? Get the hell out of here
    if (isGpsEnabled == false && isNetworkEnabled == false)
        return

    val provider = if (isGpsEnabled) LocationManager.GPS_PROVIDER else LocationManager.NETWORK_PROVIDER


    // Try last known location first
    val lastKnownLocation = locationManager.getLastKnownLocation(provider)
    if (lastKnownLocation != null) {
        val locationAgeMillis = System.currentTimeMillis() - lastKnownLocation.time
        val locationAgeSeconds = locationAgeMillis / 1000
        // Consider it stale if older than 2 minutes (120 seconds)
        if (locationAgeSeconds < 120) {
            // Location is recent enough, use it
            Location(lastKnownLocation)
            return
        }
    }

    // If last known location is null, request current location (API 30+)
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.S) {
        val locationRequest = LocationRequest.Builder(1000).setQuality(LocationRequest.QUALITY_HIGH_ACCURACY).build()
        locationManager.getCurrentLocation( provider, locationRequest,
        CancellationSignal(), Executors.newSingleThreadExecutor() ) {
            location -> loc = location
        }
    }
    else { // For older APIs, request single update as fallback
        locationManager.requestSingleUpdate(provider, { location -> callback(location) }, null)
        return
    }
}

fun getLastLocation(context: Context): Location? {
    // Check permissions
    if (ActivityCompat.checkSelfPermission(context, Manifest.permission.ACCESS_FINE_LOCATION) != PackageManager.PERMISSION_GRANTED &&
        ActivityCompat.checkSelfPermission(context, Manifest.permission.ACCESS_COARSE_LOCATION) != PackageManager.PERMISSION_GRANTED ) {
        return null
    }

    val locationManager = context.getSystemService(Context.LOCATION_SERVICE) as LocationManager

    // Check if providers are enabled
    val isGpsEnabled = locationManager.isProviderEnabled(LocationManager.GPS_PROVIDER)
    val isNetworkEnabled = locationManager.isProviderEnabled(LocationManager.NETWORK_PROVIDER)

    // No provider enabled? Get the hell out of here
    if (isGpsEnabled == false && isNetworkEnabled == false)
        return null

    val provider = if (isGpsEnabled) LocationManager.GPS_PROVIDER else LocationManager.NETWORK_PROVIDER

    // Try last known location first
    val lastKnownLocation = locationManager.getLastKnownLocation(provider)
    if (lastKnownLocation != null) {
        val locationAgeMillis = System.currentTimeMillis() - lastKnownLocation.time
        val locationAgeSeconds = locationAgeMillis / 1000
        // Consider it stale if older than 2 minutes (120 seconds)
        if (locationAgeSeconds < 120) {
            // Location is recent enough, use it
            return lastKnownLocation
        }
    }
    return null
}